#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "../vs56_shm.h"
#include "../vs56_shm_key.h"


struct TVchan *tvch[NumCh], *curch=NULL;
static int *shm_pid;

/* Allocate shared memory segment */
void get_vs56_sh_mem() {
   int i, shmid;
   unsigned char *addr;

   if(curch != NULL) return;

   /* try to find existing one */
   shmid = shmget(shm_key.code, sizeof(struct TVchan)*NumCh, 0777);
   if (shmid<0) {
      char msg[80];
      sprintf(msg,"Can't find shared memory segment '%s'",shm_key.name);
      perror(msg);
      sleep(1);
      exit(errno);
   }

   /* attach it to our memory space */
   addr = (unsigned char *)shmat ( shmid, NULL, 0 );
   if ((int)(addr) == -1) {
      char msg[80];
      sprintf(msg,"Can't attach shared memory segment '%s'",shm_key.name);
      perror(msg);
      sleep(1);
      exit(errno);
   }
   if((shmctl(shmid, SHM_LOCK, NULL) < 0) < 0) {
      char msg[80];
      sprintf(msg,"Can't prevents swapping of shared memory segment '%s'",shm_key.name);
      perror(msg);
      sleep(1);
      exit(errno);
   }
   fprintf(stderr,"Find&attach shared memory segment '%s' %dbytes at %x \n",
	   shm_key.name, sizeof(struct TVchan)*NumCh, (unsigned int)addr);

   shm_pid =  (int *)(addr + sizeof(struct TVchan)*NumCh);

   for(i=0,curch=(struct TVchan *)addr; i<NumCh; i++,curch++) {
      tvch[i] = curch;
   }
   curch = (struct TVchan *)addr;
}

void my_sleep(double dt)
{
  int nfd;
  struct timeval tv;
   tv.tv_sec = (int)dt;
   tv.tv_usec = (int)((dt - tv.tv_sec)*1000000.);
   if(tv.tv_sec==0 && tv.tv_usec<10000) tv.tv_usec=10000;
slipping:
   nfd = select(0, (fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL, &tv);
   if(nfd < 0) {
      if(errno == EINTR)
	/*On  Linux,  timeout  is  modified to reflect the amount of
	 time not slept; most other implementations DO NOT do this!*/
	 goto slipping;
      fprintf(stderr,"Error in mydelay(){ select() }. %s\n",strerror(errno));
   }
}


int vs56_ok() {
   if ( *shm_pid > 0 ) {
      if(kill(*shm_pid, 0) >= 0) return(1);
      else if(errno != ESRCH) return(1);
   }
   return(0);
}

short *tv_chan_image(int nchan) {
   if(curch == NULL)
       get_vs56_sh_mem();
   curch = tvch[nchan];
   return( curch->image );
}

double tv_chan_exp(double sec) {
   double osec;
   if(curch == NULL)
       get_vs56_sh_mem();
   if(curch->tandem) {
      osec = curch->exposure;
      if(sec>0.)
	 curch->exposure = sec;
      return (osec);
   } else
      return (0.);
}
void tv_chan_go(int clear, int nexp, int nadd, int ccd_mode) {
   if(curch == NULL)
       get_vs56_sh_mem();
   curch->command = IADD | ((clear)? ICLEAR : INULL) | (ccd_mode? ICCD : 0);
   curch->nadd = nadd;
   curch->nexp = nexp;
}
int tv_chan_ccd_mode() {
   return( (curch->command&ICCD)? 1 : 0 );
}
void tv_chan_stop() {
   if(curch == NULL)
       get_vs56_sh_mem();
   curch->command = INULL|(curch->command&ICCD);
   curch->nadd = 0;
   curch->nexp = 0;
}
void tv_chan_init() {
   if(curch == NULL)
       get_vs56_sh_mem();
   curch->command = IINIT|(curch->command&ICCD);
   curch->nadd = 0;
   curch->nexp = 0;
}

void tv_chan_clear() {
   if(curch == NULL)
       get_vs56_sh_mem();
   curch->command = ICLEAR|(curch->command&ICCD);
   curch->nadd = 0;
   curch->nexp = 0;
}

void tv_chan_write() {
   if(curch == NULL)
       get_vs56_sh_mem();
   curch->command = IWRITE|(curch->command&ICCD);
   curch->nadd = 0;
   curch->nexp = 0;
}

int tv_chan_elref(int lev) {
   int oldlev;
   if(curch == NULL)
       get_vs56_sh_mem();
   oldlev = curch->ref_lev;
   if (lev>0 && lev<64)
      curch->ref_lev = lev;
   return oldlev;
}

int tv_chan_form(int form) {
   if(curch == NULL)
       get_vs56_sh_mem();
   if ( form > 0 )
      curch->next_form = form;
   return curch->form;
}

void tv_chan_frame(int col0, int lin0, int width, int height) {
   if(curch == NULL)
       get_vs56_sh_mem();
   curch->frm_col0 = col0;
   curch->frm_lin0 = lin0;
   curch->frm_width = width;
   curch->frm_height = height;
}

void tv_chan_get_frame(int *col0, int *lin0, int *width, int *height) {
   if(curch == NULL)
       get_vs56_sh_mem();
   *col0   = curch->frm_col0;
   *lin0   = curch->frm_lin0;
   *width  = curch->frm_width;
   *height = curch->frm_height;
}

int tv_chan_img_width() {
   if(curch == NULL)
       get_vs56_sh_mem();
   return(curch->img_width);
}

int tv_chan_img_height() {
   if(curch == NULL)
       get_vs56_sh_mem();
   return(curch->img_height);
}

int tv_chan_seq_num() {
   if(curch == NULL) get_vs56_sh_mem();
   return(curch->seqn);
}

int tv_chan_set_seq_num(int new_num) {
   int old_num;
   if(curch == NULL) get_vs56_sh_mem();
   old_num = curch->seqn;
   curch->seqn = new_num;
   return(old_num);
}

char *tv_chan_msg() {
   if(curch == NULL) get_vs56_sh_mem();
   return(curch->msg);
}

int tv_chan_seq_msg() {
   if(curch == NULL) get_vs56_sh_mem();
   return(curch->seqmsg);
}

int tv_chan_tandem() {
   if(curch == NULL) get_vs56_sh_mem();
   return(curch->tandem);
}

int tv_chan_activ() {
   if(curch == NULL) get_vs56_sh_mem();
   return(curch->activ);
}

int tv_chan_tstart() {
   if(curch == NULL) get_vs56_sh_mem();
   return(curch->tstart);
}

int tv_chan_tcurr() {
   if(curch == NULL) get_vs56_sh_mem();
   return(curch->tcurr);
}

int tv_chan_added() {
   if(curch == NULL) get_vs56_sh_mem();
   return(curch->nadded);
}

int tv_chan_addlev() {
   if(curch == NULL) get_vs56_sh_mem();
   return(curch->addlev);
}

void tv_chan_set_added(int nadd) {
   if(curch == NULL) get_vs56_sh_mem();
   curch->nadded = nadd;
}

void tv_chan_set_addlev(int tv_lev) {
   if(curch == NULL) get_vs56_sh_mem();
   curch->addlev = tv_lev;
}

int tv_chan_ccd_bias(int new_bias_lev) {
   int old_one;
   if(curch == NULL)
       get_vs56_sh_mem();
   old_one = curch->bias;
   curch->bias = new_bias_lev;
   return old_one;
}

int *tv_chan_ccd_flat(int *flats, int first, int last) {
   int i;
   if(curch == NULL)
       get_vs56_sh_mem();
   if(last>MaxWidth) last=MaxWidth;
   if(first>last) first=last;
   if(flats!=NULL && last>=0) {
      for(i=first; i<last; i++)  curch->flat[i] = flats[i];
   }
   return(curch->flat);
}
