#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/times.h>


#include "vs56_io.h"
#include "vs56_shm.h"
#include "vs56_addr.h"
#include "vs56_comn.h"

int hist[256], imin=0, imax=0, hmax=0;
double hmed=0., havr=0., hdisp=0.;

void mydelay(double dt)
{
  int nfd;
  struct timeval tv;
   tv.tv_sec = (int)dt;
   tv.tv_usec = (int)((dt - tv.tv_sec)*1000000.);
   if(tv.tv_sec==0 && tv.tv_usec<10000) tv.tv_usec=10000;
slipping:
   nfd = select(0, (fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL, &tv);
   if(nfd < 0) {
      if(errno == EINTR)
	/*On  Linux,  timeout  is  modified to reflect the amount of
	 time not slept; most other implementations DO NOT do this!*/
	 goto slipping;
      fprintf(stderr,"Error in mydelay(){ select() }. %s\n",strerror(errno));
   }
}

void set_ref_lev(int high)
{
     high &= 0x3f;
     outp(adc_adr_56, 1);
     outp(0x80,0);
     outp(adc_command_56,high<<2);   /*   IOUT0   */
     outp(0x80,0);
     outp(adc_adr_56,2);
     outp(0x80,0);
     outp(adc_command_56,0);         /*   IOUT1   */
     outp(0x80,0);
     outp(adc_adr_56,0);
     outp(0x80,0);
printf("set_ref_lev(%d)\n", high);
}

const int Odd = 0;
const int Next = 1;
const int Even = 2;
int wait_switch(int odd_even) {
    int i, a=odd_even;
    if(a==Next) a=inp(rs_read_56) & 2;
    for(i=40000; (--i>0) && (a == (inp(rs_read_56) & 2)); ); /*tout~40ms*/
    return(i);
}

int wait_sync() {
    int i;
    i = wait_switch( Odd );
    if(i>0)
       i = wait_switch( Even );
    return(i);
}

int inputFrame56()
{
    int i,j,a;

    i = wait_switch( Next );
    if(i>0) {
       outp(write_56, 1|tv_out);

       mydelay(0.011); /* : mydelay(0.041);*/

       for(j=40000; (--j>0) && ((inp(rs_read_56) & 8) != 0); ); /*tout~40ms*/
       if(j==0) return 2;

       a = wait_switch( Next );
       outp(write_56, 0);
    }
/*printf("inputFrame56(): wait sync = %.3fms\n",(40000.-i)/1000.);*/
    if(i==0) return 1;
/*printf("inputFrame56(): wait end of input = %.3fms\n",(40000.-j)/1000.+10.);*/
    if(j==0) return 2;
    return 0;
}

#define INTERNAL 0
#define EXTERNAL 1
#define ACC_STOP  0
#define ACC_START 1
#define FIELD 0
#define FRAME 1
const int agc_on = 0;
const int gamma_off = 1;
const int exposure = 7;

static void setTamdem(sync_mode, acc_start, acc_type) {
    int i,code,bit1;
    code=(sync_mode&1) | ((acc_start&1) << 1) | ((acc_type&1) << 2) |
	((agc_on&1) << 3) | ((gamma_off&1) << 4) | ((exposure&7) << 5);
    for(i=0; i<8; i++) {
	bit1 = (code >> i) & 1;
	outp(out_56,bit1|0);
	outp(out_56,bit1|2);
	outp(out_56,bit1|0);
    }
  if(exposure == 7)
    outp(out_56,0);
  else
    outp(out_56,1);
}

int inputFrameTandem(double sec)
{
    int i, ret=0;
    char msg[100];

    setTamdem(INTERNAL, ACC_STOP,  FRAME);

    if((i=wait_sync())==0) {ret = 1; goto restore;}

    setTamdem(INTERNAL, ACC_START, FRAME);
    for(; sec>0.01 && curch->command&IADD; sec -= 0.04) {
       if( fmod(sec, 1.0) < 0.05) {
	  sprintf(msg,"Tandem Exposure: %.1f",sec);
	  wall_msg(msg);
       } else
	  mydelay(0.021);
       if((i=wait_sync())==0) {ret = 1; goto restore;}
    }
    setTamdem(EXTERNAL, ACC_STOP,  FRAME);
    outp(write_56, 1|tv_out);
    mydelay(0.061);
    if((i=wait_sync())==0) {ret = 1; goto restore;}
    /*for(i=40000; (--i>0) && ((inp(rs_read_56) & 8) != 0); ); tout~40ms*/
    /*if(i==0) ret = 2;*/
    outp(write_56, 0);
restore:
    setTamdem(EXTERNAL, ACC_STOP,  FIELD);
    sprintf(msg,"Tandem Exposure: %s", ret? "FAIL" : "Ok");
    wall_msg(msg);
    return ret;
}

void read_frame(int ADD, int Lev0, int CCD) {
    int i,j,c,n,n05,n50,n95,i05,i95;
    int i0 = curch->frm_lin0;
    int in = i0 + curch->frm_height;
    int ie = curch->img_height;
    int j0 = curch->frm_col0;
    int jn = j0 + curch->frm_width;
    int je = curch->img_width;
    short *p = curch->image;
    int ccdlev0 = curch->bias;
    int ccd_favr = 1;
    int *ccd_f = curch->flat;

    if(i0>=in) in=i0+1;
    if(j0>=jn) jn=j0+1;

    for(i=0; i<256; i++) hist[i]=0;
    imax = 0; imin=255;

    if(CCD) {
       Lev0 -= ccdlev0;
       for(j=j0; j<jn; j++) {
	  if(ccd_f[j]<1) ccd_f[j]=1;
	  ccd_favr += ccd_f[j];
       }
       ccd_favr = (ccd_favr + (jn-j0)/2) / (jn-j0);
    }
    outp(write_56, 0);
    outp(reset_line_56,0);

    for(n=i=0; i<ie; i++) {
       if(i>=i0 && i<in) {
	  if(CCD)
	     ccd_f = curch->flat;
	  for(j=0; j<je; j++, p++) {
	     if(j < j0) {
		c = inp(dan_read_56);
		hist[c]++;
		n++;
	     } else if(j < jn) {
		c = inp(dan_read_56);
		hist[c]++;
		n++;
		if(c<imin) imin=c;
		if(c>imax) imax=c;
		if(CCD)
		   //c = ((c-ccdlev0)*ccd_favr)/(*ccd_f);
		   c = (((c-ccdlev0)<<8)-(*ccd_f)+ccd_favr+128)>>8;
		if(ADD)
		   c = *p + (c - Lev0);
		else
		   c -= Lev0;
		if(c < -32768) c = -32768;
		if(c > 32767 ) c = 32767;
		*p = c;
	     }
	     if(CCD) ccd_f++;
	  }
       } else
	  p += je;
       outp(inc_line_56,0);
    }
    i = wait_switch( Next );
    outp(write_56, tv_out);

    n05 = (int)(n*0.05 + 0.5);
    n50 = n/2;
    n95 = (int)(n*0.95 + 0.5);

    for(hmax=i=j=0; i<256; i++) {
      if(hist[hmax] < hist[i]) hmax=i;
      c = j;
      j += hist[i];
      if(j==n50) hmed=i;
      else if(c<n50 && n50<j) hmed=i-1.+(double)(n50-c)/(j-c);
      if(c<n05 && n05<=j) i05=i;
      if(c<n95 && n95<=j) i95=i;
    }
    for(havr=0.,j=0,i=i05; i<=i95; i++) {
      havr += hist[i]*i;
      j += hist[i];
    }
    havr /= j;
    for(hdisp=0.,i=i05; i<=i95; i++) {
       double d = i-havr;
       hdisp += d*d*hist[i];
    }
    hdisp = sqrt(hdisp/j);
}

void write_image() {
    int i,j;
    int ie = curch->img_height;
    int je = curch->img_width;
    ushort *p = curch->image;

    outp(write_56, 0);
    outp(reset_line_56,0);

    for(i=0; i<ie; i++) {
       for(j=0; j<je; j++, p++)
	  outp(dan_56, *p);
       outp(inc_line_56,0);
    }
    outp(write_56, tv_out);
}

void wall_msg( char *new_msg) {
    int nch;
    for(nch=0; nch<NumCh; nch++) {
	strncpy(tvch[nch]->msg, new_msg, 99);
	tvch[nch]->seqmsg++;
    }
}
