#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sched.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>

#include "vs56_io.h"
#include "vs56_shm.h"
#include "vs56_shm_key.h"
#include "vs56_addr.h"
#include "vs56_comn.h"

int base_addr=0x300;
unsigned short base_port_56;
int ver_board_56=2;
unsigned char tv_out=2;
unsigned char level_syng_56=0xc;  /* level of syng detection bt252  */
/*for long cable set global variable level_syng_56=0x4*/
char spe_dir[100] = "spe";

struct ChanDef chan0 = {
   0, {3854, {'v','v','6','1','.','s','p','e',0}, {0}, 7302}, 768, 512,
      {3854, {'v','v','5','2','.','s','p','e',0}, {0}, 8977}, 512, 512
};
struct ChanDef chan1 = {
   1, {3854, {'v','t','5','7','6','.','s','p','e',0}, {0}, 8109 }, 768, 576,
      {3854, {'v','t','5','1','2','.','s','p','e',0}, {0}, 10593}, 512, 512
};
struct ChanDef chan2 = {
   0, {3854, {'v','v','6','1','.','s','p','e',0}, {0}, 7302}, 768, 512,
      {3854, {'v','v','5','2','.','s','p','e',0}, {0}, 8977}, 512, 512
};
struct ChanDef chan3 = {
   0, {3854, {'v','v','6','1','.','s','p','e',0}, {0}, 7302}, 768, 512,
      {3854, {'v','v','5','2','.','s','p','e',0}, {0}, 8977}, 512, 512
};
struct ChanDef *chan[NumCh] = {&chan0, &chan1, &chan2, &chan3};

struct TVchan *tvch[NumCh], *curch=NULL, *tandch=NULL;

static unsigned char *shm_addr;

#include "vs56_conf.h"

/* Allocate shared memory segment */
void get_sh_mem() {
   int i,j, shmid, new_shm=0;
   short *p;
   int shm_size = sizeof(struct TVchan)*NumCh;
   int *shm_pid, my_pid;

   if(curch != NULL) return;

   /* first try to find existing one */
   shmid = shmget(shm_key.code, shm_size+sizeof(int), 0777);
   if (shmid<0 && errno==ENOENT) {
      /* if no - try to create a new one */
      shmid = shmget(shm_key.code, shm_size+sizeof(int), IPC_CREAT|IPC_EXCL|0777);
      new_shm=1;
   }
   if (shmid<0) {
      char msg[80];
      sprintf(msg,"Can't create shared memory segment '%s'",shm_key.name);
      perror(msg);
      exit(errno);
   }

   /* attach it to our memory space */
   shm_addr = shmat ( shmid, NULL, 0 );
   if ((int)(shm_addr) == -1) {
      char msg[80];
      sprintf(msg,"Can't attach shared memory segment '%s'",shm_key.name);
      perror(msg);
      exit(errno);
   }
   if((shmctl(shmid, SHM_LOCK, NULL) < 0) < 0) {
      char msg[80];
      sprintf(msg,"Can't prevents swapping of shared memory segment '%s'",shm_key.name);
      perror(msg);
      exit(errno);
   }

   if(new_shm)
      fprintf(stderr,"Create shared memory segment '%s' %dbytes at %x \n",
	   shm_key.name, sizeof(struct TVchan)*NumCh, shm_addr);
   else
      fprintf(stderr,"Attach shared memory segment '%s' %dbytes at %x \n",
	   shm_key.name, sizeof(struct TVchan)*NumCh, shm_addr);

   my_pid = getpid();
   shm_pid =  (int *)(shm_addr + shm_size);
   if(!new_shm) {
      if( *shm_pid > 0 && kill(*shm_pid, 0) >= 0) {
	 fprintf(stderr,"vs56 daemon already running! (PID=%d)\n", *shm_pid);
	 sleep(2);
	 exit(1);
      }
   }
   *shm_pid = my_pid;

   for(i=0,curch=(struct TVchan *)shm_addr; i<NumCh; i++,curch++) {
      tvch[i] = curch;
      if(curch->tandem)
	    tandch=curch;
      if(new_shm) {
	 curch->num=i;
	 curch->tandem=chan[i]->tandem;
	 curch->activ=0;
	 curch->exposure = 0.04;
	 curch->seqn=0;
	 curch->tstart = curch->tcurr = time(0);
	 curch->command=INULL;
	 curch->nexp=0;
	 curch->nadd=0;
	 curch->nadded=0;
	 curch->addlev=0;
	 curch->next_form=1;
	 curch->form=1;
	 curch->ref_lev=60;
	 curch->img_width  = curch->frm_width  = chan[i]->cols1;
	 curch->img_height = curch->frm_height = chan[i]->lines1;
	 curch->frm_col0 = curch->frm_lin0 = 0;
	 curch->seqmsg=0;
	 curch->msg[0]='\0';
	 curch->bias = 0;
	 for(j=0; j<MaxWidth; j++)
	     curch->flat[j] = 1;
	 for(p=curch->image,j=0; j<MaxWidth*MaxHeight; j++)
	     *p++ = (j % curch->img_width) - 100;
      }
   }
   curch = (tandch)? tandch : (struct TVchan *)shm_addr;
}

void read_conf() {
   FILE *fd;
   char msg[80], *p;
   if((fd=fopen("vs56.conf","r"))==NULL) {
      sprintf(msg,"Can't read config.file: vs56.conf ");
      perror(msg);
      exit(errno);
   }
   while(fgets(msg,79,fd)) {
      char *name,*val;
      struct ConfParam *pp;
      if(msg[0] == '#')
	 continue;
      if((p=strchr(msg,'#')) != 0)
	 *p = '\0';
      if( (name=strtok(msg," \t=")) && (val=strtok(NULL," \t\n\r"))) {
	for(pp=conf; pp->name; pp++) {
	   if (!strcmp(name,pp->name)) {
	      switch(pp->type) {
		 case 'c':
		    *(pp->addr.c) = *val;
		    break;
		 case 'i':
		    *(pp->addr.i) = strtol(val,(char **)NULL, 0);
		    break;
		 case 'f':
		    *(pp->addr.f) = atof(val);
		    break;
		 case 's':
		    strcpy(pp->addr.s,val);
		    break;
		 default:
		    break;
	      }
	      break;
	   }
	}
	if(!pp->name) goto wrong_arg;
      } else {
wrong_arg: fprintf(stderr,"Wrong conf.parameter: %s\n", msg);
      }
   }
   fclose(fd);
}

void read_spe_files() {
    int i,j,n,cs,fd;
    struct ChanDef *ch;
    struct SpeDef *spe;
    char sname[200];

    for(i=0;i<4;i++) {
       ch = chan[i];
       for(j=1;j<3;j++) {
	  spe = (j==1)? &(ch->spe1) : &(ch->spe2);
	  strcpy(sname,spe_dir);
	  strcat(sname,"/");
	  strcat(sname,spe->name);
	  if((fd = open(sname, O_RDONLY)) < 0) {
	     fprintf(stderr,"Can't load SPE-file: %s\n", sname);
	     exit(1);
	  }
	  if((n=read(fd, spe->buf, SPE_MAX_SIZE)) != spe->size) {
	     fprintf(stderr,"Wrong SPE(%s) size: %d (%d needs)\n", spe->name,n,spe->size);
	     exit(1);
	  }
	  for(cs=n=0; n<spe->size; n++) cs += spe->buf[n];
	  if(cs != spe->sum) {
	     fprintf(stderr,"Wrong SPE(%s) checksum: %d (%d needs)\n", spe->name,cs,spe->sum);
	     exit(1);
	  }
	  close(fd);
       }
    }
}

static int euid,ruid;

void init_io()
{
   unsigned short port;

   ruid = getuid();      /* save real user ID */
   euid = geteuid();     /* save effective user ID (must be "root") */

   if(ruid != 0 && euid != 0) {
      fprintf(stderr, "Wrong user ID (must be 'root')! \n");
      exit(1);
   }
   setreuid(euid,euid);  /* set "full root" */

   /* get I/O permissions for I/O registers */
   if (ioperm(base_addr, 1, 1)) {
       fprintf(stderr, "Can't get I/O permissions (are you a SuperUser?)\n");
       exit (1);
   }
   for (port = 1; port < 0x20; port++)
       ioperm(base_addr+port, 1, 1);
   ioperm(0x80, 1, 1);

   base_port_56 = base_addr;

   setreuid(ruid,euid);   /* restore effective and real user IDs */
}

static void bad_addr(int sig) {
    fprintf(stderr, "Wrong VS56 BaseAddress = %x\n",base_addr);
    exit(1);
}

void reset_xil56()
{   int i;
    outp(xil_reset_56, 0);
    for(i=0;i<100;i++) outp(0x80,0);   /* delay ~100us > 5us */
    outp(xil_set_56, 0);
    for(i=0;i<1000;i++) outp(0x80,0);  /* delay ~1ms > 400us */
}

static void load_xil56(char *xil_buf, int xil_num)
{
    int i,j,b;

    reset_xil56();

    for(i=0; i<xil_num; i++) {
	b = xil_buf[i];
	for(j=0;j<8;j++)  {
	   outp(xil_dan_56, b>>j);
	   outp(0x80,0);
	}
    }
    for(j=0;j<8;j++) {
       outp(xil_dan_56, 1);
       outp(0x80,0);
    }
}

static void init_lut_outp() {
    int i;
    outp(ram_dac_adr_56,0);
    outp(0x80,0);
    for(i=0;i<256;i++) {
      outp(ram_dac_dan_56,i);
      outp(0x80,0);
      outp(ram_dac_dan_56,i);
      outp(0x80,0);
      outp(ram_dac_dan_56,i);
      outp(0x80,0);
    }
  /*outp(ram_dac_ovl_adr_56,0);
   *for(i=0;i<16;i++)  {
   *   outp(ram_dac_ovl_dan_56,0);
   *   outp(ram_dac_ovl_dan_56,0);
   *   outp(ram_dac_ovl_dan_56,0);
   *}
   */
    outp(ram_dac_mask_56,0xff);
}
void init_lut_inp()
{
    int i;
    outp(adc_adr_56,0);
    outp(0x80,0);
    for(i=0;i<256;i++) {
       outp(adc_lut_56, i);
       outp(0x80,0);
    }
}

int set_first_chan = 1;
#define first_time set_first_chan

void set_chan(int ch_num, int ch_form) {
    static int last_ch = 0;
    int prev_form = curch->form;
    int force_load = curch->tandem;
    int loaded = 0;
    struct ChanDef *chd;
    struct SpeDef *spe;
    char *xbuf, *name,*oname;
    int a,i, nx;


    ch_num &= 3;
    if (ch_form < 1) ch_form = 1;
    if (ch_form > 2) ch_form = 2;

    if(first_time) last_ch=curch->num;

    chd = chan[ch_num];
    curch->activ=0;
    curch = tvch[ch_num];
    curch->activ=1;
    if(curch->tandem) force_load=1;

    if (ch_form == 1) {
      spe =  &(chd->spe1);
      if(curch->form != 1) {
	 curch->img_width  = curch->frm_width  = chd->cols1;
	 curch->img_height = curch->frm_height = chd->lines1;
	 curch->frm_col0 = curch->frm_lin0 = 0;
      }
    }
    if (ch_form == 2) {
      spe = &(chd->spe2);
      if(curch->form != 2) {
	 curch->img_width  = curch->frm_width  = chd->cols2;
	 curch->img_height = curch->frm_height = chd->lines2;
	 curch->frm_col0 = curch->frm_lin0 = 0;
      }
    }
    curch->form = ch_form;
    xbuf = spe->buf;
    nx = spe->size;
    name = spe->name;
    if (prev_form == 1)
       oname = chan[last_ch]->spe1.name;
    else
       oname = chan[last_ch]->spe2.name;
    if (strcmp(name, oname)) force_load=1;

force_load=1;          /* now try to reload in every chan-switch! */
//set_ref_lev(curch->ref_lev);   /* set lev, and then reload XIL??? */

    if (first_time || force_load) {
       load_xil56(xbuf, nx);
       if(first_time)
	  printf("Load XIL:  %s\n",name);
       else if (strcmp(name, oname)==0)
	  printf("ReLoad XIL:  %s\n",name);
       else
	  printf("Change XIL: %s -> %s\n",oname,name);
       init_lut_inp();
       if (ver_board_56==2) init_lut_outp();
       loaded = 1;
    }
    if (loaded || last_ch != ch_num) {
       outp(adc_adr_56, 0);
       outp(0x80,0);
       outp(adc_command_56, level_syng_56|(ch_num<<4)|(ch_num<<6));
       outp(0x80,0);
printf("Change TV channel: %d -> %d\n",last_ch,ch_num);
    }

    set_ref_lev(curch->ref_lev);   /* reload XIL, and then set lev??? */
/*    set_ref_lev(curch->ref_lev);
 */
/*
 *   if (!first_time && ver_board_56==2 && last_ch != ch_num) {
 *      write_image();
printf("Write chan%d image\n", ch_num);
 *      sleep(1);
 *   }
*/
    a = inp(rs_read_56) & 2;
    for(i=100000; (--i>0) && (a == (inp(rs_read_56) & 2)); ); /*tout~0.1s*/

    outp(write_56, 0|tv_out);
    outp(0x80,0);
    last_ch = ch_num;
    first_time = 0;
}


static void vs56_check()
{
    int i,a;

    a = inp(rs_read_56) & 2;
    for(i=100000; (--i>0) && (a == (inp(rs_read_56) & 2)); ); /*tout~0.1s*/
    outp(reset_line_56,0);
    outp(dan_56,0x55);
    outp(dan_56,0xaa);
    outp(reset_line_56,0);
    if(0x55 != inp(dan_read_56) ||
       0xaa != inp(dan_read_56)) {
       fprintf(stderr, "Test of VS56 on-board memory FAILED!!!\n");
       fprintf(stderr, "Maybe no board at BaseAddress = 0x%x ?\n",base_addr);
       /*exit(1);*/
    }
    if(i==0)
       fprintf(stderr, "Warning! VS56 - no VIDEO?!\n");
}


void vs56_init() {
    static struct sched_param p;
    int pid=0;

    read_conf();
    read_spe_files();

    get_sh_mem();
    signal(SIGHUP,  vs56_exit);
    signal(SIGINT,  vs56_exit);
    signal(SIGQUIT, vs56_exit);
    signal(SIGTERM, vs56_exit);

    p.sched_priority = 1;
    if (sched_setscheduler(pid, SCHED_FIFO, &p)) {
       perror("Can't set scheduler priority");
       exit(errno);
    }
    if (mlockall(MCL_CURRENT)) {
       perror("Can't lock process memory");
       exit(errno);
    }

    init_io();

    if (ver_board_56==1) tv_out = 0;
    else if (ver_board_56==2) tv_out = 2;

    signal(SIGSEGV,bad_addr);
    set_chan(curch->num, curch->next_form);
    signal(SIGSEGV,SIG_DFL);

    vs56_check();
}

void vs56_exit(int nsig) {
   int shm_size = sizeof(struct TVchan)*NumCh;
   int *shm_pid;
   if(nsig)
      signal(nsig,SIG_IGN);
   shm_pid =  (int *)(shm_addr + shm_size);
   *shm_pid = 0;
   exit(nsig);
}
