/* x86 I/O functions for VS56 control programs under Linux      */
/*   "Byte" I/O        -  outp(), inp()                         */
/*   "Word" I/O        -  outw(), inw()                         */
/*   "string versions" - insb_rep(), insb_loop(), outsb_rep()   */
/*                       insw_rep(), insw_loop(), outsw_rep()   */

#define SLOW_DOWN_IO __asm__ __volatile__("outb %al,$0x80")

static inline void outp (unsigned short port, unsigned char val)
  {
   __asm__ volatile ("out%B0 %0,%1" : :"a" (val), "d" (port));
  }

static inline unsigned char inp (unsigned short port)
 {
   unsigned int ret;
   __asm__ volatile ("in%B0 %1,%0" : "=a" (ret) : "d" (port));
   return ret;
 }
static inline void insb_rep(unsigned short port, void * addr, unsigned long count)
{
__asm__ volatile ("cld ; rep ; insb"
	      :"=D" (addr), "=c" (count) : "d" (port),"0" (addr),"1" (count));
}

static inline void insb_loop(unsigned short port, void * addr, unsigned long count)
{
__asm__ volatile ("cld ; loopr: insb; loop loopr"
	      :"=D" (addr), "=c" (count) : "d" (port),"0" (addr),"1" (count));
}

static inline void outsb_rep(unsigned short port, void * addr, unsigned long count)
{
__asm__ volatile ("cld ; rep ; outsb"
	      :"=S" (addr), "=c" (count) : "d" (port),"0" (addr),"1" (count));
}

static inline void outw(unsigned short port, unsigned short value)
{
__asm__ volatile ("outw %0,%1"
		::"a" ((unsigned short) value),"d" ((unsigned short) port));
}

static inline unsigned short inw(unsigned short port)
{
	unsigned short _v;
__asm__ volatile ("inw %1,%0"
		:"=a" (_v):"d" ((unsigned short) port));
	return _v;
}
static inline void insw_rep(unsigned short port, void * addr, unsigned long count)
{
__asm__ volatile ("cld ; rep ; insw"
	      :"=D" (addr), "=c" (count) : "d" (port),"0" (addr),"1" (count));
}

static inline void insw_loop(unsigned short port, void * addr, unsigned long count)
{
__asm__ volatile ("cld ; loopr: insw; loop loopr"
	      :"=D" (addr), "=c" (count) : "d" (port),"0" (addr),"1" (count));
}

static inline void outsw_rep(unsigned short port, void * addr, unsigned long count)
{
__asm__ volatile ("cld ; rep ; outsw"
	      :"=S" (addr), "=c" (count) : "d" (port),"0" (addr),"1" (count));
}
